"""
PUP Plugin defining macOS directories.
"""

import pathlib
import types

from .. import dirs



class Directories:

    @staticmethod
    def usable_in(ctx):
        return (
            (ctx.pkg_platform == 'win32') and
            (ctx.tgt_platform == 'win32')
        )

    def __call__(self, ctx, _dsp):
        return {
            **dirs.DIRS,
            'cache': pathlib.Path.home() / 'AppData' / 'Local' / 'pup',
        }



"""
PUP Plugin implementing the 'mac.app-bundle-template' step.
"""

import logging
import pathlib
import shutil

import cookiecutter
from cookiecutter import generate
try:
    # Python < 3.9
    import importlib_resources as ilr
except ImportError:
    import importlib.resources as ilr

from . import app_bundle_template



_log = logging.getLogger(__name__)



class Step:

    """
    Extracts a `cookiecutter`-based macOS Application Bundle template into the
    `build` directory (template variables are sourced from the context). Sets
    the context `python_runtime_dir`, pointing to where the Python runtime
    should be copied.
    """

    @staticmethod
    def usable_in(ctx):
        return (
            (ctx.pkg_platform == 'darwin') and
            (ctx.tgt_platform == 'darwin')
        )

    def __call__(self, ctx, dsp):

        build_dir = dsp.directories()['build']
        build_dir.mkdir(parents=True, exist_ok=True)

        tmpl_path = ilr.files(app_bundle_template)
        tmpl_data = {
            'cookiecutter': {
                'app_bundle_name': ctx.nice_name,
                'bundle_identifier': ctx.application_id,
                'version_string': ctx.src_metadata.version,
                'copyright': self._copyright_from_context(ctx),
                'launcher_name': ctx.nice_name,
                'tcl_library': str(ctx.python_rel_tcl_library),
                'python_rel_exe': str(ctx.python_rel_exe),
                'launch_module': self._launch_module_from_context(ctx),
            }
        }

        # "Generate + Remove + Generate" motivation: cookiecutter either fails
        # if the output path exists, or overwrites it. However, it does not
        # remove pre-existing files that are no longer templated. Thus, the
        # "proper" way to ensure output is consistent without deleting the
        # whole build directory is to "Generate + Remove + Generate again".

        result_path = generate.generate_files(tmpl_path, tmpl_data, build_dir, overwrite_if_exists=True)
        shutil.rmtree(result_path, ignore_errors=True)
        result_path = generate.generate_files(tmpl_path, tmpl_data, build_dir)

        # Remove the .gitignore file in template (keeps empty dir in git).
        (pathlib.Path(result_path) / 'Contents/Resources/.gitignore').unlink()

        # Copy any given icon file into the bundle.
        if ctx.icon_path:
            shutil.copyfile(
                ctx.icon_path,
                pathlib.Path(result_path) / 'Contents/Resources/Icon.icns'
            )

        new_python_runtime_dir = pathlib.Path(result_path) / 'Contents/Resources/Python'
        ctx.python_runtime_dir.replace(new_python_runtime_dir)
        ctx.python_runtime_dir = new_python_runtime_dir


    def _copyright_from_context(self, ctx):

        author = ctx.src_metadata.author
        license = ctx.src_metadata.license

        return f'{author}, {license}'


    def _launch_module_from_context(self, ctx):

        return ctx.launch_module if ctx.launch_module else ctx.src_metadata.name



volume_name = '{{cookiecutter.volume_name}}'
filename = '{{cookiecutter.filename}}'

{% if cookiecutter.icon_path %}
icon = '{{cookiecutter.icon_path}}'
{% endif %}

app_bundle = '{{cookiecutter.app_bundle_name}}'

files = [
    f'../{app_bundle}',
]
symlinks = {
    'Applications': '/Applications',
}

default_view = 'icon-view'
background = 'builtin-arrow'

text_size = 14
icon_size = 96

icon_locations = {
    'Applications': (500, 120),
    app_bundle: (140, 120),
}

{% if cookiecutter.license_path %}
license = {
    'default-language': 'en_US',
    'licenses': {
         'en_US': '{{ cookiecutter.license_path }}',
     },
}
{% endif %}



"""
PUP Plugin implementing the 'pup.metadata' step.
"""

import logging
import os
import subprocess
import sys
import tempfile

import pkginfo


_log = logging.getLogger(__name__)



class Step:

    """
    Populates the context with source package metadata.
    """

    # Strategy
    # --------
    # Use `https://pypi.org/project/pkginfo/` to collect non-installed package
    # metadata. Its API depends on the "source kind" (sdist, wheel, etc.), so:
    #
    # - Run `pip wheel` to create a wheel file from the source.
    # - Use pkginfo wheel API to collect source package metadata.

    @staticmethod
    def usable_in(ctx):
        # Should work "everywhere".
        return True

    _METADATA_FIELDS = [
        'author',
        'author_email',
        'classifiers',
        'description',
        'description_content_type',
        'download_url',
        'home_page',
        'keywords',
        'license',
        'maintainer',
        'maintainer_email',
        'metadata_version',
        'name',
        'obsoletes',
        'obsoletes_dist',
        'platforms',
        'project_urls',
        'provides',
        'provides_dist',
        'provides_extras',
        'requires',
        'requires_dist',
        'requires_external',
        'requires_python',
        'summary',
        'supported_platforms',
        'version',
    ]

    def __call__(self, ctx, dsp):

        src = ctx.src
        _log.info('Collecting metadata for %r.', src)

        with tempfile.TemporaryDirectory(prefix='pup-metadata-') as temp_dir:
            wheel_file = self._create_wheel(ctx.src, temp_dir, dsp)
            ctx.src_metadata = pkginfo.Wheel(wheel_file)

        for field in self._METADATA_FIELDS:
            _log.debug('%s=%r', field, getattr(ctx.src_metadata, field))


    def _create_wheel(self, src, work_dir, dsp):

        src_abs = os.path.abspath(src)
        cmd = [sys.executable, '-m', 'pip', 'wheel', '--no-deps', src_abs]
        
        _log.info('About to run %r.', ' '.join(cmd))

        cwd = os.getcwd()
        try:
            os.chdir(work_dir)
            dsp.spawn(
                cmd,
                out_callable=lambda line: _log.info('pip out: %s', line),
                err_callable=lambda line: _log.info('pip err: %s', line),
            )
            wheel_file = os.listdir()[0]
        finally:
            os.chdir(cwd)

        return os.path.join(work_dir, wheel_file)



"""
PUP Plugin defining macOS packaging stages.
"""


class Steps:

    @staticmethod
    def usable_in(ctx):
        return (
            (ctx.pkg_platform == 'darwin') and
            (ctx.tgt_platform == 'darwin')
        )

    def __call__(self, ctx, _dsp):
        return (
            'pup.python-runtime',
            'mac.app-bundle-template',
            'pup.pip-install',
            'pup.install-cleanup',
            'mac.sign-app-bundle',
            'mac.notarize-app-bundle',
            'mac.create-dmg',
        )



"""
Command line interface for PUP.
"""

import functools
import logging
import os
import sys

import click

from . import __title__ as sw_title, __version__ as sw_version
from . import api
from . import logs



_log = logging.getLogger(__package__)



def command_wrapper(command_function):
    """
    Decorator for click.command functions that i) logs version at start/finish,
    ii) logs exceptions raised during execution, iii) sets process exit status.
    """
    @functools.wraps(command_function)
    def wrapper(*args, **kw):
        _log.info('%s %s - starting with PID=%r', sw_title, sw_version, os.getpid())
        try:
            exit_code = command_function(*args, **kw)
        except Exception as exc:
            _log.critical('Execution failure: %s', exc)
            _log.critical('Traceback below:', exc_info=exc)
            exit_code = -1
        except BaseException:
            # KeyboardInterrupt exceptions (CTRL-C, SIGINT)
            exit_code = -2
        finally:
            _log.info('%s %s - done', sw_title, sw_version)
            sys.exit(exit_code)

    return wrapper




@click.group()
@click.version_option(version=sw_version)
@click.option(
    '--log-level',
    type=click.Choice(
        ['CRITICAL', 'ERROR', 'WARNING', 'WARN', 'INFO', 'DEBUG'],
        case_sensitive=False,
    ),
    default='INFO',
    show_default=True,
    envvar='PUP_LOG_LEVEL',
)
def main(log_level):
    """
    Pluggable Micro Packager.
    """
    logs.start(log_level)



@main.command()
@click.option('--ignore-plugin', 'ignore_plugins', multiple=True)
@click.option('--launch-module')
@click.option('--nice-name')
@click.option('--icon-path')
@click.option('--license-path')
@click.argument('src')
@command_wrapper
def package(src, ignore_plugins, launch_module, nice_name, icon_path, license_path):
    """
    Packages the GUI application in the given pip-installable source.
    """
    return api.package(
        src,
        ignore_plugins=ignore_plugins,
        launch_module=launch_module,
        nice_name=nice_name,
        icon_path=icon_path,
        license_path=license_path,
    )



"""
Pluggable Micro Packager
"""

__version__ = '1.0.0a10'

__title__ = 'pup'
__description__ = 'Pluggable Micro Packager'

__license__ = 'MIT'
__uri__ = 'https://github.com/mu-editor/pup/'

__author__ = 'Tiago Montes'
__email__ = 'tiago.montes@gmail.com'



from . api import package

__all__ = ['package']



"""
PUP Plugin implementing the 'pup.install-cleanup' step.
"""

import logging
import shutil


_log = logging.getLogger(__name__)



class Step:


    @staticmethod
    def usable_in(ctx):
        return (
            (ctx.pkg_platform == 'darwin') or
            (ctx.pkg_platform == 'win32')
        ) and (
            (ctx.pkg_platform == ctx.tgt_platform)
        )

    def __call__(self, ctx, dsp):

        self._delete_unneeded_scripts(ctx)
        self._delete_unneeded_paths(ctx)
        self._delete_unneeded_files(ctx)
        self._compile_lib(ctx, dsp)


    def _delete_unneeded_scripts(self, ctx):

        python_exe = ctx.python_runtime_dir / ctx.python_rel_exe
        python_scripts_path =  ctx.python_runtime_dir / ctx.python_rel_scripts

        for file_path in python_scripts_path.glob('*'):
            if file_path == python_exe:
                continue
            if file_path.is_dir():
                shutil.rmtree(str(file_path), ignore_errors=True)
            elif not file_path.is_symlink() or file_path.resolve() != python_exe:
                # Delete any file that doesn't symlink to the Python executable.
                file_path.unlink()


    def _delete_unneeded_paths(self, ctx):

        paths_to_keep = {
            ctx.python_runtime_dir / rel_path
            for rel_path in (
                ctx.python_rel_stdlib,
                ctx.python_rel_site_packages,
                ctx.python_rel_scripts,
                'DLLs',
                'Tcl',
            )
        }

        for candidate in ctx.python_runtime_dir.iterdir():
            candidate_path = ctx.python_runtime_dir / candidate
            if not candidate_path.is_dir():
                continue
            if candidate_path in paths_to_keep:
                continue
            # Don't delete if candidate_path is parent of any of paths_to_keep.
            if any(candidate_path == parent for p in paths_to_keep for parent in p.parents):
                continue
            _log.info('Deleting %r...', str(candidate_path))
            shutil.rmtree(
                str(candidate_path),
                ignore_errors=True,
            )


    _GLOBS_TO_DELETE = (
        '*.lib',
        '*.pdb',
        '*.a',
    )

    def _delete_unneeded_files(self, ctx):

        for glob in self._GLOBS_TO_DELETE:
            _log.info('Deleting %r files...', glob)
            for file_path in ctx.python_runtime_dir.glob(f'**/{glob}'):
                file_path.unlink()


    def _compile_lib(self, ctx, dsp):

        # Compiles both the Standard Library and Site Packages.

        python_exe = ctx.python_runtime_dir / ctx.python_rel_exe
        python_stdlib = ctx.python_runtime_dir / ctx.python_rel_stdlib

        compile_cmd = [
            str(python_exe),
            '-m',
            'compileall',
            '-f',
            '-q',
            str(python_stdlib),
        ]

        _log.info('Compiling the Standard Library and site-packages...')
        dsp.spawn(
            compile_cmd,
            out_callable=lambda line: _log.info('compile out: %s', line),
            err_callable=lambda line: _log.info('compile err: %s', line),
        )



"""
PUP Plugin implementing the 'pup.pip-install' step.
"""

import logging
import subprocess


_log = logging.getLogger(__name__)



class Step:

    """
    Downloads 
    """

    @staticmethod
    def usable_in(ctx):
        return (
            (ctx.pkg_platform == 'darwin') or
            (ctx.pkg_platform == 'win32')
        ) and (
            (ctx.pkg_platform == ctx.tgt_platform)
        )

    def __call__(self, ctx, dsp):

        cmd = [
            str(ctx.python_runtime_dir / ctx.python_rel_exe),
            '-m',
            'pip',
            'install',
            '--no-warn-script-location',
            ctx.src,
        ]

        _log.info('About to run %r.', ' '.join(cmd))

        dsp.spawn(
            cmd,
            out_callable=lambda line: _log.info('pip out: %s', line),
            err_callable=lambda line: _log.info('pip err: %s', line),
        )



"""
PUP Plugin implementing the 'win.sign-binaries' step.
"""

import logging
import os
import pathlib
import platform



_log = logging.getLogger(__name__)



class _Signer:

    """
    Signs executables and shared libraries to be distributed
    using Windows SDK's signtool.exe.
    """

    @staticmethod
    def usable_in(ctx):
        return (
            (ctx.pkg_platform == 'win32') and
            (ctx.tgt_platform == 'win32')
        )


    def __init__(self):

        self._identity = None
        self._signtool = None


    def __call__(self, ctx, dsp):

        try:
            self._identity = os.environ['PUP_SIGNING_IDENTITY']
        except KeyError as exc:
            raise RuntimeError(f'environment variable {exc} not defined') from None

        self._signtool = self._find_signtool(ctx, dsp)


    _SIGNTOOL_DIR_FROM_PYTHON_ARCH = {
        '32bit': 'x86',
        '64bit': 'x64',
    }

    def _find_signtool(self, ctx, dsp):

        # AFAICT there's no "central registry" (pun intended!)
        # where the available `signtool.exe` can be located.

        # We'll look for it under the directories defined by
        # the Windows PROGRAMFILES* environment variables.
        program_files = [
            v for k, v in os.environ.items()
            if k.upper().startswith('PROGRAMFILES')
        ]

        # Expected to be under a "../bin/<version>/" directory.
        signtool_paths = [
            path
            for pf in program_files
            for path in pathlib.Path(pf).glob('**/bin/**/signtool.exe')
        ]
        _log.debug('signtool_paths=%r', signtool_paths)

        # Figure out which, if any, apply to our architecture.
        arch, _ = platform.architecture()
        try:
            signtool_dir = self._SIGNTOOL_DIR_FROM_PYTHON_ARCH[arch]
        except KeyError:
            raise RuntimeError(f'no signtool.exe found for arch={arch!r}')

        # Sort candidates so that we can get the "most recent" one.
        signtool_paths = sorted(
            path for path in signtool_paths
            if path.parent.name.lower() == signtool_dir.lower()
        )
        _log.debug('signtool_paths=%r', signtool_paths)
        
        try:
            signtool_path = signtool_paths[-1]
        except IndexError:
            raise RuntimeError('no usable signtool.exe found')

        return signtool_path


    def sign_one_file(self, dsp, path):

        cmd = [
            str(self._signtool),
            'sign',
            '/q',
            '/n', self._identity,
            '/fd', 'SHA256',
            # TODO: Do not hardcode this URL, grab it from an env var.
            '/tr', 'http://timestamp.digicert.com',
            '/td', 'SHA256',
            str(path)
        ]
        _log.info('Signing %r...', str(path))
        dsp.spawn(
            cmd,
            out_callable=lambda line: _log.info('signtool.exe out: %s', line),
            err_callable=lambda line: _log.info('signtool.exe err: %s', line),
        )



class SignBinaries(_Signer):

    def __call__(self, ctx, dsp):

        try:
            super().__call__(ctx, dsp)
        except RuntimeError as exc:
            _log.error('cannot sign: %s.', exc)
            return

        self._sign_binaries(ctx, dsp)


    def _sign_binaries(self, ctx, dsp):

        for extension in ('exe', 'dll', 'pyd'):
            for path in ctx.relocatable_root.glob(f'**/*.{extension}'):
                self.sign_one_file(dsp, path)



class SignMSI(_Signer):

    def __call__(self, ctx, dsp):

        try:
            super().__call__(ctx, dsp)
        except RuntimeError as exc:
            _log.error('cannot sign: %s.', exc)
            return

        self.sign_one_file(dsp, ctx.final_artifact.absolute())



"""
PUP Plugin implementing the 'pup.python-runtime' step.
"""

import json
import logging
import os
import pathlib
import shutil
import tarfile
import tempfile as tf

import zstandard


_log = logging.getLogger(__name__)



_PYTHON_BUILD_STANDALONE_URLs = {
    'darwin': {
        '3.7': 'https://github.com/indygreg/python-build-standalone/releases/download/20200823/cpython-3.7.9-x86_64-apple-darwin-pgo-20200823T2228.tar.zst',
        '3.8': 'https://github.com/indygreg/python-build-standalone/releases/download/20200823/cpython-3.8.5-x86_64-apple-darwin-pgo-20200823T2228.tar.zst',
    },
    'win32': {
        '3.7': 'https://github.com/indygreg/python-build-standalone/releases/download/20200822/cpython-3.7.9-x86_64-pc-windows-msvc-shared-pgo-20200823T0118.tar.zst',
        '3.8': 'https://github.com/indygreg/python-build-standalone/releases/download/20200830/cpython-3.8.5-x86_64-pc-windows-msvc-shared-pgo-20200830T2254.tar.zst',
    },
}



class Step:

    """
    Downloads and extracts a relocatable Python runtime for the target platform
    into the directory specified by the context.
    """

    @staticmethod
    def usable_in(ctx):
        return (
            (ctx.pkg_platform == 'darwin') or
            (ctx.pkg_platform == 'win32')
        ) and (
            (ctx.pkg_platform == ctx.tgt_platform)
        )

    def __call__(self, ctx, dsp):

        build_dir = self._ensure_build_dir(dsp)
        pbs_dir = build_dir / 'pbs'

        pbs_artifact = self._get_pbs_artifact(ctx, dsp)
        self._extract_zstd_file(pbs_artifact, pbs_dir)

        pbs_py_dir = pbs_dir / 'python'
        pbs_py_json = self._load_pbs_python_json(pbs_py_dir / 'PYTHON.json')

        pbs_py_paths = pbs_py_json['python_paths']
        pbs_data = pbs_py_paths['data']

        # Delete unneeded things.
        self._delete_test_packages(pbs_py_dir, pbs_py_json)
        self._delete_platform_config(pbs_py_dir, pbs_py_json)

        # Track key Python Runtime paths.
        ctx.python_runtime_dir = pbs_py_dir / pbs_data
        ctx.python_rel_exe = self._relative(pbs_py_json['python_exe'], pbs_data)
        ctx.python_rel_scripts = self._relative(pbs_py_paths['scripts'], pbs_data)
        ctx.python_rel_stdlib = self._relative(pbs_py_paths['stdlib'], pbs_data)
        ctx.python_rel_site_packages = self._relative(pbs_py_paths['purelib'], pbs_data)

        # Find `init.tcl` path (macOS needs the TCL_LIBRARY env var set).
        tcl_library_path = pbs_py_dir / pbs_py_json['tcl_library_path']
        init_tcl_path = None
        for candidate_path in pbs_py_json['tcl_library_paths']:
            init_tcl_path = tcl_library_path / candidate_path / 'init.tcl'
            if init_tcl_path.exists():
                break
        ctx.python_rel_tcl_library = self._relative(
            self._relative(init_tcl_path.parent, pbs_py_dir),
            pbs_data,
        )


    def _ensure_build_dir(self, dsp):

        build_dir = dsp.directories()['build']
        build_dir.mkdir(parents=True, exist_ok=True)
        return build_dir


    def _get_pbs_artifact(self, ctx, dsp):

        url = os.environ.get(
            'PUP_PBS_URL',
            self._pbs_url(ctx.tgt_platform, ctx.tgt_python_version_suffix)
        )
        return dsp.download(url)


    def _pbs_url(self, platform, py_version):

        try:
            return _PYTHON_BUILD_STANDALONE_URLs[platform][py_version]
        except KeyError:
            raise RuntimeError(f'No {platform} Python runtime for Python {py_version}.')


    def _extract_zstd_file(self, filename, target_dir):

        with open(filename, 'rb') as input_file:
            decompressor = zstandard.ZstdDecompressor()
            with decompressor.stream_reader(input_file) as reader:
                with tarfile.open(mode='r|', fileobj=reader) as tf:
                    tf.extractall(path=target_dir)


    def _load_pbs_python_json(self, filename):

        with open(filename, 'rt', encoding='utf8') as input_file:
            return json.load(input_file)

    def _delete_tree(self, path):

        delete = str(path)
        _log.debug('Deleting %r...', delete)
        shutil.rmtree(delete, ignore_errors=True)


    def _delete_test_packages(self, pbs_py_dir, pbs_py_json):

        _log.info('Deleting Standard Library test packages...')
        stdlib_path = pbs_py_json['python_paths']['stdlib']
        for test_package in pbs_py_json['python_stdlib_test_packages']:
            test_package_path = test_package.replace('.', '/')
            self._delete_tree(pbs_py_dir / stdlib_path / test_package_path)


    def _delete_platform_config(self, pbs_py_dir, pbs_py_json):

        _log.info('Deleting Platform Config files...')
        stdlib_platform_config = pbs_py_json.get('python_stdlib_platform_config')
        if not stdlib_platform_config:
            return
        self._delete_tree(pbs_py_dir / stdlib_platform_config)


    def _relative(self, path, base):

        return pathlib.Path(path).relative_to(base) if path else None



"""
Pluggable Micro Packager API.
"""

import logging
import sys

from . import context
from . import dispatcher



_log = logging.getLogger(__name__)
_log.addHandler(logging.NullHandler())



def _context(
    ignore_plugins,
    src=None,
    launch_module=None,
    nice_name=None,
    icon_path=None,
    license_path=None,
):

    return context.Context(
        src=src,
        launch_module=launch_module,
        nice_name=nice_name,
        icon_path=icon_path,
        license_path=license_path,
        ignore_plugins=ignore_plugins,
        platform=sys.platform,
        python_version=sys.version_info,
    )



def package(
    src,
    *,
    ignore_plugins=(),
    launch_module=None,
    nice_name=None,
    icon_path=None,
    license_path=None,
):

    _log.info('Package %r: starting.', src)

    ctx = _context(ignore_plugins, src, launch_module, nice_name, icon_path, license_path)
    dsp = dispatcher.Dispatcher(ctx)

    dsp.collect_src_metadata()

    for step in dsp.steps():
        _log.info('Step %r: starting.', step)
        dsp.run_pluggable_step(step)
        _log.info('Step %r: completed.', step)

    _log.info('Package %r: completed.', src)



def directories(*, ignore_plugins=()):

    ctx = _context(ignore_plugins)
    dsp = dispatcher.Dispatcher(ctx)

    return dsp.directories()



def download(url, *, ignore_plugins=()):

    ctx = _context(ignore_plugins)
    dsp = dispatcher.Dispatcher(ctx)

    return dsp.download(url)



"""
PUP Plugin implementing the 'win.create-msi' step.
"""

import logging
import os
import pathlib
import re
import shutil
import uuid
import zipfile

import cookiecutter
from cookiecutter import generate
try:
    # Python < 3.9
    import importlib_resources as ilr
except ImportError:
    import importlib.resources as ilr

from . import msi_wxs_template



_log = logging.getLogger(__name__)



class Step:

    """
    Extracts a `cookiecutter`-based Windows distribution template into the
    `build` directory (template variables are sourced from the context). Sets
    the context `python_runtime_dir`, pointing to where the Python runtime
    should be copied.
    """

    @staticmethod
    def usable_in(ctx):
        return (
            (ctx.pkg_platform == 'win32') and
            (ctx.tgt_platform == 'win32')
        )

    def __call__(self, ctx, dsp):

        wix_root = self._ensure_wix(dsp)
        wix_src_path = self._create_wix_source(ctx, dsp)
        self._create_wix_manifest(ctx, dsp, wix_root, wix_src_path)
        self._compile_wix_sources(ctx, dsp, wix_root, wix_src_path)
        msi_file_path = self._link_wix_objects(ctx, dsp, wix_root, wix_src_path)

        ctx.final_artifact = msi_file_path
        _log.info('MSI file at %r.', str(msi_file_path))


    _WIX_BINARIES_URL = (
        'https://github.com/wixtoolset/wix3'
        '/releases/download/wix3112rtm/wix311-binaries.zip'
    )

    def _ensure_wix(self, dsp):

        wix_bin_zip = dsp.download(self._WIX_BINARIES_URL)
        wix_extract_dir = pathlib.Path(wix_bin_zip).with_suffix('.extracted')

        if wix_extract_dir.exists():
            return wix_extract_dir

        wix_extract_dir.mkdir()
        with zipfile.ZipFile(wix_bin_zip) as zf:
            zf.extractall(path=wix_extract_dir)

        return wix_extract_dir


    def _create_wix_source(self, ctx, dsp):

        build_dir = dsp.directories()['build']

        rtf_license_path = self._create_rtf_from_text(ctx.license_path, build_dir)

        tmpl_path = ilr.files(msi_wxs_template)
        tmpl_data = {
            'cookiecutter': {
                'app_name': ctx.nice_name,
                'version': ctx.src_metadata.version,
                'msi_version': self._msi_version(ctx.src_metadata.version),
                'icon_path': ctx.icon_path,
                'rtf_license_path': rtf_license_path,
                'author': ctx.src_metadata.author,
                'author_email': ctx.src_metadata.author_email,
                'url': ctx.src_metadata.home_page,
                'launch_module': self._launch_module_from_context(ctx),
                'guid': self._upgrade_code_guid(ctx),
            }
        }

        # "Generate + Remove + Generate" motivation: cookiecutter either fails
        # if the output path exists, or overwrites it. However, it does not
        # remove pre-existing files that are no longer templated. Thus, the
        # "proper" way to ensure output is consistent without deleting the
        # whole build directory is to "Generate + Remove + Generate again".

        result_path = generate.generate_files(tmpl_path, tmpl_data, build_dir, overwrite_if_exists=True)
        shutil.rmtree(result_path, ignore_errors=True)
        result_path = generate.generate_files(tmpl_path, tmpl_data, build_dir)

        return pathlib.Path(result_path)


    def _create_rtf_from_text(self, license_path, build_dir):

        # Horrible hack inspired by https://stackoverflow.com/questions/10725438/converting-txt-to-rtf

        if not license_path:
            return None

        rtf_path = build_dir / license_path.with_suffix('.rtf').name

        try:
            with open(rtf_path, 'wt', encoding='ASCII') as dst:
                dst.write(r'{\rtf1\ansi\pard')
                dst.write('\n')
                with open(license_path, 'rt', encoding='ASCII') as src:
                    for line in src:
                        dst.write(
                            line.replace(
                                '\\',
                                '\\\\',
                            ).replace(
                                '{',
                                '\{'
                            ).replace(
                                '}',
                                '\}',
                            )
                        )
                        dst.write('\n')
                        dst.write(r'\par')
                        dst.write('\n')
                dst.write(r'}')
        except UnicodeDecodeError:
            _log.error('could not read %r as ASCII', str(license_path))
            return None
        except OSError as exc:
            _log.error('could not convert %r to RTF: %s', str(license_path), exc)
            return None

        return rtf_path


    # Copied from PEP 440
    _VERSION_PATTERN = r"""
        v?
        (?:
            (?:(?P<epoch>[0-9]+)!)?                           # epoch
            (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
            (?P<pre>                                          # pre-release
                [-_\.]?
                (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
                [-_\.]?
                (?P<pre_n>[0-9]+)?
            )?
            (?P<post>                                         # post release
                (?:-(?P<post_n1>[0-9]+))
                |
                (?:
                    [-_\.]?
                    (?P<post_l>post|rev|r)
                    [-_\.]?
                    (?P<post_n2>[0-9]+)?
                )
            )?
            (?P<dev>                                          # dev release
                [-_\.]?
                (?P<dev_l>dev)
                [-_\.]?
                (?P<dev_n>[0-9]+)?
            )?
        )
        (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
    """

    _VERSION_RE = re.compile(
        r"^\s*" + _VERSION_PATTERN + r"\s*$",
        re.VERBOSE | re.IGNORECASE,
    )

    def _msi_version(self, version):

        # MSI versions are not as flexible as PEP 440's.
        # Let's adapt the version to three dot-separated numbers.

        result = self._VERSION_RE.match(version)
        pep440_release = result.group('release')

        numbers_only = pep440_release == version
        nums = pep440_release.split('.')
        num_count = len(nums)

        if num_count < 3:
            nums.extend(('0', '0'))

        msi_version = '.'.join(nums[:3])

        if not numbers_only or num_count != 3:
            _log.warning(
                'Version %r not MSI supported: using %r.',
                version,
                msi_version,
            )

        return msi_version


    def _launch_module_from_context(self, ctx):

        return ctx.launch_module if ctx.launch_module else ctx.src_metadata.name


    def _upgrade_code_guid(self, ctx):

        return str(uuid.uuid5(uuid.NAMESPACE_URL, ctx.src_metadata.home_page))


    def _create_wix_manifest(self, ctx, dsp, wix_root, wix_src_path):

        launch_module = self._launch_module_from_context(ctx)
        wix_manifest_path = wix_src_path / 'manifest.wxs'
        cmd = [
            str(wix_root / 'heat.exe'),
            'dir',
            str(ctx.relocatable_root),
            '-nologo',
            '-gg',
            '-sfrag',
            '-sreg',
            '-srd',
            '-scom',
            '-dr', f'{launch_module}_ROOTDIR',
            '-cg', f'{launch_module}_COMPONENTS',
            '-var', 'var.SourceDir',
            '-out', str(wix_manifest_path),
        ]

        dsp.spawn(
            cmd,
            out_callable=lambda line: _log.info('wix heat out: %s', line),
            err_callable=lambda line: _log.info('wix heat err: %s', line),
        )


    def _compile_wix_sources(self, ctx, dsp, wix_root, wix_src_path):

        # Must change CWD because candle.exe outputs to it. :/
        cwd = os.getcwd()
        try:
            os.chdir(str(wix_src_path))

            cmd = [
                str(wix_root / 'candle.exe'),
                '-nologo',
                f'-dSourceDir={ctx.relocatable_root}',
            ]
            cmd.extend(str(wxs_path) for wxs_path in pathlib.Path().glob('*.wxs'))

            dsp.spawn(
                cmd,
                out_callable=lambda line: _log.info('wix candle out: %s', line),
                err_callable=lambda line: _log.info('wix candle err: %s', line),
            )
        finally:
            os.chdir(cwd)


    def _link_wix_objects(self, ctx, dsp, wix_root, wix_src_path):

        dist_dir = dsp.directories()['dist']
        msi_file_path = dist_dir / self._msi_filename(ctx)

        cmd = [
            str(wix_root / 'light.exe'),
            '-nologo',
            '-ext', 'WixUIExtension',
            '-spdb',
            '-o', str(msi_file_path),
        ]
        cmd.extend(str(wxs_path) for wxs_path in wix_src_path.glob('*.wixobj'))

        dsp.spawn(
            cmd,
            out_callable=lambda line: _log.info('wix light out: %s', line),
            err_callable=lambda line: _log.info('wix light err: %s', line),
        )

        return msi_file_path


    def _msi_filename(self, ctx):

        return f'{ctx.nice_name} {ctx.src_metadata.version}.msi'



"""
Spawning child processes while tracking their output.
"""

import logging
import subprocess
import threading


_log = logging.getLogger(__name__)



def _copy_lines(readable, write_callable):

    for line in iter(readable.readline, ''):
        write_callable(line.rstrip('\n'))



def spawn(command, out_callable=None, err_callable=None, encoding='utf8'):
    """
    Spawns a child process specified by `command` which is passed to
    subprocess.Popen where `out/err_callable` is called for each stdout/err
    line that the child process produces, assumed to be text in the given
    `encoding`.
    """

    _log.debug('Spawning %r.', command)
    child = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        bufsize=0,
        encoding=encoding,
        universal_newlines=True,
    )
    _log.debug('Spawned PID %s.', child.pid)

    # Let's have callables to ensure we consume the child's out/err streams.
    out_callable = out_callable if out_callable else lambda line: None
    err_callable = err_callable if err_callable else lambda line: None

    # Spawn one thread per stream to drive the child to callable I/O.
    out_thread = threading.Thread(target=_copy_lines, args=(child.stdout, out_callable))
    err_thread = threading.Thread(target=_copy_lines, args=(child.stderr, err_callable))
    out_thread.start()
    err_thread.start()

    child_exit_status = child.wait()

    out_thread.join()
    err_thread.join()

    _log.debug('Child terminated with exit status %s', child_exit_status)
    return child_exit_status



"""
PUP Plugin implementing the 'mac.sign-app-bundle' step.
"""

import logging
import os
import pathlib
import shutil
import tempfile

try:
    # Python < 3.9
    import importlib_resources as ilr
except ImportError:
    import importlib.resources as ilr

from . import sign_resources



_log = logging.getLogger(__name__)



class Step:

    @staticmethod
    def usable_in(ctx):
        return (
            (ctx.pkg_platform == 'darwin') and
            (ctx.tgt_platform == 'darwin')
        )


    def __init__(self):

        self._identity = None
        self._codesign = None
        self._entitlements = None

    def __call__(self, ctx, dsp):

        try:
            self._identity = os.environ['PUP_SIGNING_IDENTITY']
        except KeyError as exc:
            _log.error('Cannot sign: environment variable %s not defined.', str(exc))
            return

        build_dir = dsp.directories()['build']
        app_bundle_name = ctx.nice_name
        app_bundle_path = build_dir / f'{app_bundle_name}.app'

        binaries_dir = (ctx.python_runtime_dir / ctx.python_rel_exe).parent

        self._entitlements = ilr.files(sign_resources) / 'entitlements.plist'
        self._codesign = shutil.which('codesign')

        self._sign_binaries(dsp, binaries_dir)
        self._sign_libraries(dsp, app_bundle_path)
        self._sign(dsp, app_bundle_path)

        self._assess_signing_result(dsp, app_bundle_path)


    def _sign_binaries(self, dsp, binaries_dir):

        for file_path in binaries_dir.glob('*'):
            if file_path.is_file() and not file_path.is_symlink():
                self._sign(dsp, file_path)


    def _sign_libraries(self, dsp, base_path):

        self._sign_framework(dsp, base_path)
        self._sign_shared_libs(dsp, base_path)
        self._sign_bundled_zips(dsp, base_path)
        self._sign_bundled_zips(dsp, base_path, extension='whl')


    def _sign_framework(self, dsp, base_path):

        for framework_path in base_path.glob('**/*.framework'):
            for file_path in framework_path.glob('**/*'):
                if file_path.is_file() and not file_path.is_symlink():
                    self._sign(dsp, file_path)


    def _sign_shared_libs(self, dsp, base_path):

        for glob in ('*.so', '*.dylib'):
            for shlib_path in base_path.glob(f'**/{glob}'):
                self._sign(dsp, shlib_path)


    def _sign_bundled_zips(self, dsp, base_path, extension='zip'):

        for zip_path in base_path.glob(f'**/*.{extension}'):
            self._sign_zip_contents(dsp, zip_path)


    def _sign_zip_contents(self, dsp, zip_path):

        _log.info('Signing %r contents...', str(zip_path))
        build_dir = dsp.directories()['build']
        with tempfile.TemporaryDirectory(prefix='sign-zip-', dir=build_dir) as td:
            dsp.spawn([
                shutil.which('unzip'),
                str(zip_path),
                '-d',
                td,
            ])
            self._sign_libraries(dsp, pathlib.Path(td))
            cwd = os.getcwd()
            # Must get absolute path before os.chdir.
            zip_path_absolute = zip_path.absolute()
            try:
                os.chdir(td)
                dsp.spawn([
                    shutil.which('zip'),
                    '-qyr',
                    str(zip_path_absolute),
                    '.',
                ])
            finally:
                os.chdir(cwd)


    def _sign(self, dsp, target):

        cmd = [
            self._codesign,
            '--sign',
            self._identity,
            '--entitlements',
            str(self._entitlements),
            '--deep',
            str(target),
            '--force',
            '--options',
            'runtime',
            '--timestamp',
        ]
        _log.info('Signing %r...', str(target))
        dsp.spawn(
            cmd,
            out_callable=lambda line: _log.info('codesign| %s', line),
            err_callable=lambda line: _log.info('codesign! %s', line),
        )


    def _assess_signing_result(self, dsp, app_bundle_path):

        cmd = [
            shutil.which('spctl'),
            '--assess',
            '-vvvv',
            str(app_bundle_path),
        ]
        _log.info('Assessing signing result...')
        dsp.spawn(
            cmd,
            out_callable=lambda line: _log.info('spctl| %s', line),
            err_callable=lambda line: _log.info('spctl! %s', line),
        )

        # TODO: Expect `spctl` output to be along the lines of this?
        #       build/pup/<app_bundle_name>: rejected
        #       source=Unnotarized Developer ID
        #       origin=<signing certificate cn>



"""
PUP Plugin defining Windows packaging stages.
"""


class Steps:

    @staticmethod
    def usable_in(ctx):
        return (
            (ctx.pkg_platform == 'win32') and
            (ctx.tgt_platform == 'win32')
        )

    def __call__(self, ctx, _dsp):
        return (
            'pup.python-runtime',
            'win.distribution-layout',
            'pup.pip-install',
            'pup.install-cleanup',
            'win.sign-binaries',
            'win.create-msi',
            'win.sign-msi',
        )



"""
PUP Plugin implementing the 'mac.notarize-app-bundle' step.
"""

import contextlib
import logging
import os
import shutil
import time
import xml.etree.ElementTree as et



_log = logging.getLogger(__name__)



class Step:

    @staticmethod
    def usable_in(ctx):
        return (
            (ctx.pkg_platform == 'darwin') and
            (ctx.tgt_platform == 'darwin')
        )


    def __call__(self, ctx, dsp):

        try:
            user = os.environ['PUP_NOTARIZE_USER']
            password = os.environ['PUP_NOTARIZE_PASSWORD']
        except KeyError as exc:
            _log.error('Cannot notarize: environment variable %s not defined.', str(exc))
        else:
            self._notarize(ctx, dsp, user, password)


    def _notarize(self, ctx, dsp, user, password):

        build_dir = dsp.directories()['build']
        app_bundle_name = ctx.nice_name
        app_bundle_path = build_dir / f'{app_bundle_name}.app'

        app_bundle_zip = self._create_app_bundle_zip(dsp, app_bundle_path)
        request_uuid = self._request_notarization(ctx, dsp, app_bundle_zip, user, password)
        self._wait_notarization(dsp, request_uuid, user, password)
        self._staple_app_bundle(dsp, app_bundle_path)
        self._assess_notarization_result(dsp, app_bundle_path)
        app_bundle_zip.unlink()


    @contextlib.contextmanager
    def _working_directory(self, directory):

        cwd = os.getcwd()
        os.chdir(directory)
        try:
            yield
        finally:
            os.chdir(cwd)


    def _create_app_bundle_zip(self, dsp, app_bundle_path):

        with self._working_directory(app_bundle_path.parent):
            dsp.spawn(
                command=[
                    shutil.which('zip'),
                    '-qyr',
                    f'{app_bundle_path.name}.zip',
                    str(app_bundle_path.name),
                ]
            )

        return app_bundle_path.with_suffix('.app.zip')


    def _parse_xml_plist(self, xml_payload):
        """
        Returns a `dict` built from parsing the XML Property List in `xml_payload`.
        """
        root_element = et.fromstring(xml_payload)
        xml_dict, *tail = root_element.findall('./dict')
        if tail:
            _log.warning('Multiple dicts in XML plist: %r', xml_payload)
        return self._parse_xml_dict(xml_dict)


    def _parse_xml_dict(self, element):
        """
        Returns a `dict` built from recursively parsing an XML element tree
        starting with a <dict> element, containing key+value element pairs:
        keys are the text within <key> elements, values are the text within
        the following elements (if such elements are <dict>, the function is
        called recursively).
        """
        result = {}
        element_iter = iter(element)
        for k_element in element_iter:
            v_element = next(element_iter)
            k = k_element.text
            v = self._parse_xml_dict(v_element) if v_element.tag == 'dict' else v_element.text
            result[k] = v
        return result


    def _request_notarization(self, ctx, dsp, app_bundle_zip, user, password):

        _log.info('Submitting notarization request...')
        cmd = [
            shutil.which('xcrun'),
            'altool',
            '--notarize-app',
            '--primary-bundle-id',
            ctx.application_id,
            '--username',
            user,
            '--password',
            password,
            '--output-format',
            'xml',
            '--file',
            app_bundle_zip,
        ]
        xml_output_lines = []
        dsp.spawn(
            cmd,
            out_callable=xml_output_lines.append,
            err_callable=lambda line: _log.info('altool! %s', line),
        )
        xml_payload = '\n'.join(xml_output_lines)
        _log.debug('xml_payload=%r', xml_payload)
        response = self._parse_xml_plist(xml_payload)
        try:
            request_uuid = response['notarization-upload']['RequestUUID']
        except KeyError as exc:
            _log.critical('Missing %s key in notarization request output.', str(exc))
            request_uuid = None

        _log.info('Notarization request submitted: RequestUUID=%r.', request_uuid)
        return request_uuid


    def _wait_notarization(self, dsp, request_uuid, user, password, delay=60, tolerance=600):

        cmd = [
            shutil.which('xcrun'),
            'altool',
            '--notarization-info',
            request_uuid,
            '--username',
            user,
            '--password',
            password,
            '--output-format',
            'xml',
        ]
        start = time.time()
        while True:
            _log.info('Requesting notarization info in %ss...', delay)
            time.sleep(delay)
            xml_output_lines = []
            dsp.spawn(
                cmd,
                out_callable=xml_output_lines.append,
                err_callable=lambda line: _log.info('altool! %s', line),
            )
            xml_payload = '\n'.join(xml_output_lines)
            _log.debug('xml_payload=%r', xml_payload)
            response = self._parse_xml_plist(xml_payload)
            response_info = response['notarization-info']
            try:
                status = response_info['Status']
            except KeyError as exc:
                _log.warning('Missing %s key in notarization info output.', str(exc))
            else:
                _log.info('Status is %r.', status)
            if status != 'in progress':
                break
            if time.time() - start > tolerance:
                _log.warning('Slow notarization: see https://developer.apple.com/system-status/')

        status_code = response_info['Status Code']
        status_msg = response_info['Status Message']
        log_url = response_info['LogFileURL']
        if status != 'success':
            _log.error(
                'Notarization failed: %r. Check log at %r.',
                status_msg,
                log_url,
            )
        elif status_code != '0':
            _log.warning(
                'Unexpected notarization status code %r: %r. Check log at %r.',
                status_code,
                status_msg,
                log_url,
            )


    def _staple_app_bundle(self, dsp, app_bundle_path):

        _log.info('Stapling...')
        cmd = [
            shutil.which('xcrun'),
            'stapler',
            'staple',
            app_bundle_path.name,
        ]
        with self._working_directory(app_bundle_path.parent):
            dsp.spawn(
                cmd,
                out_callable=lambda line: _log.info('stapler| %s', line),
                err_callable=lambda line: _log.info('stapler! %s', line),
            )


    def _assess_notarization_result(self, dsp, app_bundle_path):

        _log.info('Assessing notarization result...')
        cmd = [
            shutil.which('spctl'),
            '--assess',
            '-vvvv',
            str(app_bundle_path),
        ]
        dsp.spawn(
            cmd,
            out_callable=lambda line: _log.info('spctl| %s', line),
            err_callable=lambda line: _log.info('spctl! %s', line),
        )

        # TODO: Expect `spctl` output to be along the lines of this?
        #       build/pup/<app_bundle_name>: accepted
        #       source=Notarized Developer ID
        #       origin=<signing certificate cn>



"""
Packaging Context.
"""

import pathlib
from urllib import parse


class Context:

    def __init__(
        self,
        *,
        src,
        launch_module,
        nice_name,
        icon_path,
        license_path,
        ignore_plugins,
        platform,
        python_version,
    ):

        self.src = src
        self.launch_module = launch_module
        self._nice_name = nice_name
        self.icon_path = pathlib.Path(icon_path).absolute() if icon_path else None
        self.license_path = pathlib.Path(license_path).absolute() if license_path else None

        if icon_path and not self.icon_path.exists():
            raise EnvironmentError(f'Non-existent icon path {icon_path!r}.')

        if license_path and not self.license_path.exists():
            raise EnvironmentError(f'Non-existent license path {license_path!r}.')

        self.src_metadata = None

        self.ignore_plugins = ignore_plugins
        self.pkg_platform = platform
        self.tgt_platform = platform
        self.tgt_python_version = python_version[:3]
        self.tgt_python_version_suffix = '.'.join(map(str, python_version[:2]))

        self.relocatable_root = None

        self.python_runtime_dir = None
        self.python_rel_exe = None
        self.python_rel_scripts = None
        self.python_rel_stdlib = None
        self.python_rel_site_packages = None
        self.python_rel_tcl_library = None

        self.final_artifact = None


    @property
    def nice_name(self):
        """
        User facing packaged name.
        """
        return self._nice_name or self.src_metadata.name


    @property
    def application_id(self):
        """
        Returns the application identifier built from the package's home_page
        URL, consisting of two-sets of '.'-separated strings: the reverse DNS
        host/domain part, followed by the in-order path components.

        Example:
        - home_page='https://example.com/a/path'
        - bundle_id='com.example.a.path'
        """

        url_parts = parse.urlsplit(self.src_metadata.home_page)
        return '.'.join((
            '.'.join(reversed(url_parts.netloc.split('.'))),
            '.'.join(filter(None, url_parts.path.split('/')))
        ))



"""
PUP Plugin implementing the 'mac.create-dmg' step.
"""

import logging
import os
import pathlib
import shutil
import sysconfig

import cookiecutter
from cookiecutter import generate
try:
    # Python < 3.9
    import importlib_resources as ilr
except ImportError:
    import importlib.resources as ilr

from . import dmgbuild_settings_template



_log = logging.getLogger(__name__)



class Step:

    """
    Uses `dmgbuild` to create a DMG image of the application bundle.
    """

    @staticmethod
    def usable_in(ctx):
        return (
            (ctx.pkg_platform == 'darwin') and
            (ctx.tgt_platform == 'darwin')
        )

    def __call__(self, ctx, dsp):

        volume_name = f'{ctx.nice_name} {ctx.src_metadata.version}'
        filename = f'{volume_name}.dmg'

        settings_path = self._create_dmgbuild_settings(ctx, dsp, volume_name, filename)
        self._create_dmg_file(ctx, dsp, settings_path)
        self._deliver_dmg_file(dsp, settings_path, filename)


    def _create_dmgbuild_settings(self, ctx, dsp, volume_name, filename):

        tmpl_path = ilr.files(dmgbuild_settings_template)
        tmpl_data = {
            'cookiecutter': {
                'app_name': ctx.nice_name,
                'volume_name': volume_name,
                'filename': filename,
                'icon_path': ctx.icon_path,
                'license_path': ctx.license_path,
                'app_bundle_name': f'{ctx.nice_name}.app',
            }
        }

        # "Generate + Remove + Generate" motivation: cookiecutter either fails
        # if the output path exists, or overwrites it. However, it does not
        # remove pre-existing files that are no longer templated. Thus, the
        # "proper" way to ensure output is consistent without deleting the
        # whole build directory is to "Generate + Remove + Generate again".

        build_dir = dsp.directories()['build']
        result_path = generate.generate_files(tmpl_path, tmpl_data, build_dir, overwrite_if_exists=True)
        shutil.rmtree(result_path, ignore_errors=True)
        result_path = generate.generate_files(tmpl_path, tmpl_data, build_dir)

        return pathlib.Path(result_path)


    def _create_dmg_file(self, ctx, dsp, settings_path):

        # The `dmgbuild` settings file includes relative paths that require
        # us launching it with its directory as the CWD.

        cwd = os.getcwd()
        try:
            os.chdir(str(settings_path))

            cmd = [
                os.path.join(sysconfig.get_path('scripts'), 'dmgbuild'),
                '-s', 'settings.py',
                # Both overridden in the settings file.
                'volume-name',
                'output.dmg',
            ]
            dsp.spawn(
                cmd,
                out_callable=lambda line: _log.info('dmgbuild out: %s', line),
                err_callable=lambda line: _log.info('dmgbuild err: %s', line),
            )
        finally:
            os.chdir(cwd)


    def _deliver_dmg_file(self, dsp, path, filename):

        dist_dir = dsp.directories()['dist']
        dist_dir.mkdir(exist_ok=True)

        try:
            # Move file to file instead of file to dir ensures overwrites;
            # otherwise, move fails if file exists on destination dir.
            shutil.move(
                path / filename,
                dist_dir / filename,
            )
        except OSError as exc:
            _log.error('%s', exc)



import io
import os
import re

import setuptools



###############################################################################

NAME = "pup"
META_PATH = os.path.join("src", "pup", "__init__.py")
KEYWORDS = ["packaging", "gui", "applications"]
CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Environment :: Console",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Natural Language :: English",
    "Operating System :: MacOS",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX :: Linux",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Topic :: Software Development :: Libraries :: Python Modules",
    "Topic :: System :: Installation/Setup",
    "Topic :: System :: Software Distribution",
]
INSTALL_REQUIRES = [
    "importlib-metadata==3.3.0;python_version<'3.8'",
    "click==7.1.2",
    "httpx==0.16.1",
    "wheel==0.36.2",
    "pkginfo==1.6.1",
    "importlib-resources==4.1.1;python_version<'3.9'",
    "cookiecutter==1.7.2",
    "zstandard==0.15.1",
    "dmgbuild==1.4.2;sys_platform=='darwin'",
]
EXTRAS_REQUIRE = {
    "docs": [
        "sphinx",
        "towncrier",
    ],
    "tests": [
        "coverage",
    ],
    "release": [
        "twine",
    ],
}
EXTRAS_REQUIRE["dev"] = EXTRAS_REQUIRE["tests"] + EXTRAS_REQUIRE["docs"]



###############################################################################

HERE = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    """
    Build an absolute path from *parts* and and return the contents of the
    resulting file.  Assume UTF-8 encoding.
    """
    with io.open(os.path.join(HERE, *parts), encoding="utf-8") as f:
        return f.read()



META_FILE = read(META_PATH)

def find_meta(meta):
    """
    Extract __*meta*__ from META_FILE.
    """
    meta_match = re.search(
        r"^__{meta}__ = ['\"]([^'\"]*)['\"]".format(meta=meta),
        META_FILE, re.M
    )
    if meta_match:
        return meta_match.group(1)
    raise RuntimeError("Unable to find __{meta}__ string.".format(meta=meta))



if __name__ == "__main__":
    setuptools.setup(
        name=NAME,
        description=find_meta("description"),
        license=find_meta("license"),
        url=find_meta("uri"),
        version=find_meta("version"),
        author=find_meta("author"),
        author_email=find_meta("email"),
        maintainer=find_meta("author"),
        maintainer_email=find_meta("email"),
        keywords=KEYWORDS,
        long_description=read("README.rst"),
        long_description_content_type='text/x-rst',
        packages=setuptools.find_packages(where="src"),
        package_dir={"": "src"},
        zip_safe=False,
        classifiers=CLASSIFIERS,
        install_requires=INSTALL_REQUIRES,
        extras_require=EXTRAS_REQUIRE,
        entry_points={                                                                    
            'console_scripts': [
                'pup=pup.__main__:main',
            ],
            'pup.plugins': [
                'pup.metadata=pup.plugins.metadata:Step',
                'pup.download=pup.plugins.download:Step',
                'pup.python_runtime=pup.plugins.python_runtime:Step',
                'pup.pip_install=pup.plugins.pip_install:Step',
                'pup.install_cleanup=pup.plugins.install_cleanup:Step',
                'mac.dirs=pup.plugins.mac.dirs:Directories',
                'mac.steps=pup.plugins.mac.steps:Steps',
                'mac.app_bundle_template=pup.plugins.mac.app_bundle:Step',
                'mac.sign_app_bundle=pup.plugins.mac.sign:Step',
                'mac.notarize_app_bundle=pup.plugins.mac.notarize:Step',
                'mac.create_dmg=pup.plugins.mac.create_dmg:Step',
                'win.dirs=pup.plugins.win.dirs:Directories',
                'win.steps=pup.plugins.win.steps:Steps',
                'win.distribution_layout=pup.plugins.win.dist_layout:Step',
                'win.create_msi=pup.plugins.win.create_msi:Step',
                'win.sign_binaries=pup.plugins.win.signer:SignBinaries',
                'win.sign_msi=pup.plugins.win.signer:SignMSI',
            ],
        },
        include_package_data=True,
    )




"""
PUP logging setup module.
"""

import os
import logging
import sys



def _level_from_string(log_level_str):

    return getattr(logging, log_level_str.upper())



_OUTPUT_TO_TTY_FORMAT = {
    False: '%(asctime)s %(levelname).1s %(name)s %(message)s',
    True: '%(levelname).1s %(message)s',
}

def start(log_level_str):
    """
    Starts logging messages of severity `log_level_str` or higher to STDERR.
    Logged messages include more detail when STDERR is not a TTY.
    """
    output_is_tty = os.isatty(sys.stderr.fileno())
    logging.basicConfig(
        level=_level_from_string(log_level_str),
        format=_OUTPUT_TO_TTY_FORMAT[output_is_tty],
        datefmt='%Y%m%d %H%M%S',
    )
    logging.captureWarnings(True)



"""
PUP Plugin implementing the 'download' API.
"""

import hashlib
import logging

import httpx



class Step:

    """
    Downloads files from the given URL. Caches them to avoid future downloads.
    """

    @staticmethod
    def usable_in(ctx):
        return True


    def __call__(self, ctx, dsp, url):

        url_sha256 = hashlib.sha256(url.encode('utf8')).hexdigest()

        dirs = dsp.directories()
        cache_dir = dirs['cache']
        cached_download = cache_dir / url_sha256

        if cached_download.exists():
            return cached_download

        return self._download(url, dirname=cache_dir, filename=url_sha256)


    def _download(self, url, dirname, filename):

        if not dirname.exists():
            dirname.mkdir()

        file = dirname / filename

        with open(file.with_suffix('.url'), 'wt', encoding='utf8') as f:
            f.write(url)

        with open(file, 'wb') as f, httpx.stream('GET', url) as r:
            for chunk in r.iter_bytes():
                f.write(chunk)

        return file



"""
Dispatches calls to plugins.
"""

import logging

try:
    import importlib.metadata as ilm
except ImportError:
    # Python <= 3.8
    import importlib_metadata as ilm

from . import spawn



_log = logging.getLogger(__name__)



class Dispatcher:

    def __init__(self, ctx):

        self._ctx = ctx

        all_plugin_entry_points = ilm.entry_points().get('pup.plugins', ())
        self._plugin_entry_points = [
            entry_point
            for entry_point in all_plugin_entry_points
            if not self._ignore(entry_point, self._ctx.ignore_plugins)
        ]
        _log.debug('plugin_entry_points=%r', self._plugin_entry_points)


    @staticmethod
    def _ignore(entry_point, ignore_plugins):

        return any(
            entry_point.value.startswith(ignore_plugin)
            for ignore_plugin in ignore_plugins
        )


    def _classes_and_names_for(self, what):

        return [
            (entry_point.load(), entry_point.value)
            for entry_point in self._plugin_entry_points
            if entry_point.name.endswith(what)
        ]


    def _invoke_plugin(self, name, **kwargs):

        classes_and_names = [
            (plugin_class, name)
            for plugin_class, name in self._classes_and_names_for(name)
            if plugin_class.usable_in(self._ctx)
        ]
        count = len(classes_and_names)
        if count != 1:
            names = ', '.join(repr(name) for _, name in classes_and_names)
            raise RuntimeError(f'{count} {name!r} plugins: {names}.')

        plugin_class, _ = classes_and_names[0]
        plugin = plugin_class()
        return plugin(self._ctx, self, **kwargs)


    def steps(self):

        return self._invoke_plugin('.steps')


    def run_pluggable_step(self, name):

        name = name.replace('-', '_')
        return self._invoke_plugin(name)


    def directories(self):

        return self._invoke_plugin('.dirs')


    def download(self, url):

        return self._invoke_plugin('pup.download', url=url)


    def collect_src_metadata(self):

        return self._invoke_plugin('pup.metadata')


    def spawn(self, command, out_callable=None, err_callable=None, encoding='utf8'):

        return spawn.spawn(command, out_callable, err_callable, encoding)
